// ---------------------------------------------------------------------------
//	M88 - PC-8801 Emulator.
//	Copyright (C) cisc 1997, 1999.
// ---------------------------------------------------------------------------
//	$Id: wincore.h,v 1.19 1999/08/16 06:22:53 cisc Exp $

#if !defined(win32_wincore_h)
#define win32_wincore_h

// ---------------------------------------------------------------------------

#include "types.h"
#include "CritSect.h"
#include "pc88/pc88.h"
#include "pc88/config.h"
#include "winsound.h"
#include "winjoy.h"
#include "winmouse.h"

namespace PC8801
{
class WinKeyIF;
}

// ---------------------------------------------------------------------------

class WinCore : public PC88
{
public:
	WinCore();
	~WinCore();
	bool Init(WinUI* ui, HWND hwnd, Draw* draw, DiskManager* diskmgr, PC8801::WinKeyIF* keyb); 
	bool Cleanup();

	void Reset();
	void ApplyConfig(PC8801::Config* config);
	
	long GetExecCount();
	void Wait(bool dowait);
	void SetGUIFlag(bool flag);
	void ActivateMouse(bool flag);
	bool EnableCPU1Dump(bool flag);
	
	bool SaveShapshot(const char* filename);
	bool LoadShapshot(const char* filename, const char* diskname = 0);

	PC8801::WinSound* GetSound() { return &sound; }

private:
	bool ConnectDevices(PC8801::WinKeyIF* keyb);
	void Execute(long clock, long time);
	bool EnablePad(bool enable);
	bool EnableMouse(bool enable);

	uint32 GetTime();
	
	uint ThreadMain();
	static uint __stdcall ThreadEntry(LPVOID arg);
	
	CriticalSection cs_cpu;
	WinUI* ui;

	HANDLE hthread;
	uint idthread;

	int clock;
	int execcount;
	
	volatile bool shouldterminate;
	volatile bool execute;
	bool padenable;
	bool mouseenable;
	int guicount;

	PC8801::WinSound sound;
	PC8801::WinJoyPad pad;
	PC8801::WinMouse mouse;
	PC8801::Config config;
};

// ---------------------------------------------------------------------------
//	}EXgp
//
inline void WinCore::ActivateMouse(bool flag)
{
	mouse.Enable(flag && mouseenable && !guicount);
}


#endif // !defined(win32_wincore_h)
