// ---------------------------------------------------------------------------
//	M88 - PC-8801 Emulator
//	Copyright (C) cisc 1998, 1999.
// ---------------------------------------------------------------------------
//	$Id: winjoy.cpp,v 1.3 1999/07/28 15:17:04 cisc Exp $

// #define	DIRECTINPUT_VERSION		0x0300

#include "headers.h"
#include "WinJoy.h"
#include "pc88/config.h"
#include "status.h"

using namespace PC8801;

// ---------------------------------------------------------------------------
//	\z/
//
WinJoyPad::WinJoyPad()
: Device(0)
{
	enabled = false;
	button1 = JOY_BUTTON1 | JOY_BUTTON3;
	button2 = JOY_BUTTON2 | JOY_BUTTON4;
}

WinJoyPad::~WinJoyPad()
{

}

// ---------------------------------------------------------------------------
//	
//
bool WinJoyPad::Init()
{
	paravalid = false;
	enabled = false;
	if (!joyGetNumDevs())
	{
		statusdisplay.Show(70, 3000, "WCXeBbN API gpł܂");
		return false;
	}

	JOYINFO joyinfo;
	if (joyGetPos(JOYSTICKID1, &joyinfo) == JOYERR_UNPLUGGED)
	{
		statusdisplay.Show(70, 3000, "WCXeBbNڑĂ܂");
		return false;
	}
	enabled = true;
	return true;
}

// ---------------------------------------------------------------------------
//	
//
uint WinJoyPad::GetDirection(uint)
{
	if (!paravalid)
		Update();
	return data[0];
}

uint WinJoyPad::GetButton(uint)
{
	if (!paravalid)
		Update();
	return data[1];
}

// ---------------------------------------------------------------------------
//	WCXeBbN̏ԂXV
//
void WinJoyPad::Update()
{
	const int threshold = 16384;
	JOYINFO joyinfo;
	if (enabled && joyGetPos(JOYSTICKID1, &joyinfo) == JOYERR_NOERROR)
	{
		data[0] =  (joyinfo.wYpos < (32768-threshold) ? 0 : 1)	// U
				 | (joyinfo.wYpos > (32768+threshold) ? 0 : 2)	// D
				 | (joyinfo.wXpos < (32768-threshold) ? 0 : 4)	// L
				 | (joyinfo.wXpos > (32768+threshold) ? 0 : 8)	// R
				 | 0xf0;
		data[1] =  (joyinfo.wButtons & button1 ? 0 : 1)
				 | (joyinfo.wButtons & button2 ? 0 : 2)
				 | 0xfc;
	}
	else
		data[0] = data[1] = 0xff;
	paravalid = true;
}

// ---------------------------------------------------------------------------
//	VSync ݂
//
void WinJoyPad::VSync(uint, uint d)
{
	if (d)
		paravalid = false;
}
	
// ---------------------------------------------------------------------------
//	RtBM[VXV
//
void WinJoyPad::ApplyConfig(const Config* config)
{
	if (config->flags & Config::swappadbuttons)
	{
		button1 = JOY_BUTTON2 | JOY_BUTTON4;
		button2 = JOY_BUTTON1 | JOY_BUTTON3;
	}
	else
	{
		button1 = JOY_BUTTON1 | JOY_BUTTON3;
		button2 = JOY_BUTTON2 | JOY_BUTTON4;
	}
}

// ---------------------------------------------------------------------------
//	device description
//
const Device::Descriptor WinJoyPad::descriptor = { indef, outdef };

const Device::OutFuncPtr WinJoyPad::outdef[] = 
{
	STATIC_CAST(Device::OutFuncPtr, VSync),
};

const Device::InFuncPtr WinJoyPad::indef[] = 
{
	STATIC_CAST(Device::InFuncPtr, GetDirection),
	STATIC_CAST(Device::InFuncPtr, GetButton),
};

